from jaaql.openapi.swagger_documentation import *
from jaaql.constants import *
from jaaql.documentation.documentation_shared import ARG_RES__username, EXAMPLE__password, ARG_RES__password, RES__oauth_token,\
    rename_arg, set_required, set_nullable, ARG_RES__application, ARG_RES__configuration

TITLE = "JAAQL API"
DESCRIPTION = "Collection of methods in the JAAQL API"
FILENAME = "jaaql_api"

# Not unused. Used to generate html files
from jaaql.documentation.documentation_shared import DOCUMENTATION__oauth_token, DOCUMENTATION__oauth_refresh

ARG_RES__application_nullable = SwaggerArgumentResponse(
    name=KEY__application,
    description="The application, if you want to make this a public user",
    example=["playground"],
    required=False,
    arg_type=str,
    condition="If this is a public user"
)
ARG_RES__configuration_nullable = SwaggerArgumentResponse(
    name=KEY__configuration,
    description="The configuration, if you want to make this a public user",
    example=["main"],
    required=False,
    arg_type=str,
    condition="If this is a public user"
)
ARG_RES__parameters = SwaggerArgumentResponse(
    name=KEY__parameters,
    description="Nonspecific data which is supplied as an object either for email or signup. Is validated",
    arg_type=ARG_RESP__allow_all,
    required=False,
    condition="Is signup data provided"
)

DOCUMENTATION__create_account = SwaggerDocumentation(
    tags="Admin",
    methods=SwaggerMethod(
        method=REST__POST,
        name="Create account",
        description="Will create an account, if you have privileges to do so",
        body=[
            ARG_RES__username,
            ARG_RES__application_nullable,
            ARG_RES__configuration_nullable,
            SwaggerArgumentResponse(
                name=KEY__password,
                description="The password, if you want to make this a public user",
                example=EXAMPLE__password,
                required=False,
                strip=True,
                arg_type=str,
                condition="If this is a public user"
            )
        ]
    )
)

ARG_RES__email_template = SwaggerArgumentResponse(
    name=KEY__email_template,
    description="Email template name",
    arg_type=str,
    example=["signup"]
)
ARG_RES__already_signed_up_email_template = rename_arg(ARG_RES__email_template, KEY__already_signed_up_email_template,
                                                       "The email template sent if the user already exists")

ARG_RES__reset_password_email_template = rename_arg(ARG_RES__email_template, KEY__email_template,
                                                    "The email template for resetting the password")

UUID__invite = "137adf50-76a3-4314-b933-b94e6686489e"

ARG_RES__invite_key = SwaggerArgumentResponse(
    name=KEY__invite_key,
    description="A sign up key that functions as an invite for a specific email address",
    arg_type=str,
    example=[UUID__invite],
)

ARG_RES__invite_poll_key = rename_arg(ARG_RES__invite_key,
                                      new_description="A key returned after signing up that can be used to sign up to the server once the signup "
                                                      "process has began with the other key or the 5 digit signup code has been input")

ARG_RES__reset_key = rename_arg(ARG_RES__invite_key, KEY__reset_key,
                                new_description="A key that can be used to reset a password of a specific user")

ARG_RES__reset_poll_key = rename_arg(ARG_RES__reset_key,
                                     new_description="A key returned after requesting a reset password email that can be used to reset "
                                     "password once the process has began with the other key or the 8 length reset code has been input")

ARG_RES__email = rename_arg(ARG_RES__username, KEY__email)

DOCUMENTATION__sign_up_request_invite = SwaggerDocumentation(
    tags="Signup",
    security=False,
    methods=SwaggerMethod(
        name="Request an invitation",
        description="Requests an invitation, sending an email to the user (or if the email template does not contain "
        "the signup id, they will then need to be approved by an admin).",
        method=REST__POST,
        body=[
            ARG_RES__email,
            ARG_RES__parameters,
            ARG_RES__configuration,
            ARG_RES__application,
            ARG_RES__email_template,
            ARG_RES__already_signed_up_email_template
        ],
        response=[
            SwaggerResponse(
                description="Sign up response",
                response=ARG_RES__invite_poll_key
            ),
            SwaggerFlatResponse(
                description=ERR__too_many_signup_attempts,
                code=HTTPStatus.TOO_MANY_REQUESTS,
                body=ERR__too_many_signup_attempts
            )
        ]
    )
)

DOCUMENTATION__sign_up_status = SwaggerDocumentation(
    tags="Signup",
    # The security is in the invite key. User has not signed up yet so cannot get an oauth token
    security=False,
    methods=SwaggerMethod(
        name="Request invite status",
        description="Requesting status with the invite key sent to the email will allow the poll key to be used as an invite key",
        method=REST__GET,
        arguments=[
            SwaggerArgumentResponse(
                name=KEY__invite_or_poll_key,
                description="Either an invite or invite poll key",
                arg_type=str,
                example=[UUID__invite]
            ),
            SwaggerArgumentResponse(
                name=KEY__invite_code,
                description="An 5 length invite code received only via email. Used only with the invite poll key and valid only for 15 minutes. "
                "The invite key in the email is valid much longer (2 weeks default) but due to the ease of guessing a 5 letter code (~13M "
                "combinations) it is kept short for security reasons",
                arg_type=str,
                example=["AGZSB"],
                required=False,
                condition="Is this the invite poll key"
            )
        ],
        response=[
            SwaggerResponse(
                description="Invite status enumeration",
                response=SwaggerArgumentResponse(
                    name=KEY__invite_key_status,
                    description="An enumeration of the statuses of an invite key. 0->New signup, 1->Sign up process started, "
                    "2->Signing up again with same email, 3->Sign up already finished",
                    arg_type=int,
                    example=[0]
                )
            ),
            SwaggerFlatResponse(
                description=ERR__too_many_code_attempts,
                code=HTTPStatus.LOCKED,
                body=ERR__too_many_code_attempts
            )
        ]
    )
)

DOCUMENTATION__sign_up_with_invite = SwaggerDocumentation(
    tags="Signup",
    # The security is in the invite key. User has not signed up yet so cannot get an oauth token
    security=False,
    methods=SwaggerMethod(
        name="Signup with invite",
        description="Signs up to JAAQL using either the key fetched either from internal methods or from the email. "
        "Returns the signup parameters if supplied upon signup",
        method=REST__POST,
        body=[
            ARG_RES__invite_key,
            ARG_RES__password
        ],
        response=[
            SwaggerResponse(
                description="Information relating to the sign up",
                response=ARG_RES__email
            ),
            SwaggerFlatResponse(
                description=ERR__already_signed_up,
                code=HTTPStatus.CONFLICT,
                body=ERR__already_signed_up
            )
        ]
    )
)

DOCUMENTATION__sign_up_fetch = SwaggerDocumentation(
    tags="Signup",
    methods=SwaggerMethod(
        name="Fetch signup data",
        description="Fetches the signup data",
        method=REST__GET,
        arguments=ARG_RES__invite_key,
        response=[
            SwaggerResponse(
                description="Finish signup response",
                response=ARG_RES__parameters
            )
        ]
    )
)

DOCUMENTATION__sign_up_finish = SwaggerDocumentation(
    tags="Signup",
    methods=SwaggerMethod(
        name="Finish signup",
        description="Deletes the signup data",
        method=REST__POST,
        body=ARG_RES__invite_key
    )
)

DOCUMENTATION__reset_password = SwaggerDocumentation(
    security=False,
    tags="Account",
    methods=SwaggerMethod(
        name="Reset Password",
        description="Send reset password email",
        method=REST__POST,
        body=[
            ARG_RES__email,
            ARG_RES__reset_password_email_template,
            ARG_RES__application,
            ARG_RES__configuration,
            ARG_RES__parameters
        ],
        response=[
            SwaggerResponse(
                description="Sign up response",
                response=ARG_RES__reset_poll_key
            ),
            SwaggerFlatResponse(
                description=ERR__too_many_reset_requests,
                code=HTTPStatus.TOO_MANY_REQUESTS,
                body=ERR__too_many_signup_attempts
            )
        ]
    )
)

DOCUMENTATION__reset_password_status = SwaggerDocumentation(
    tags="Account",
    # The security is in the invite key. User has not signed up yet so cannot get an oauth token
    security=False,
    methods=SwaggerMethod(
        name="Request password reset status",
        description="Requesting status with the reset key sent to the email will allow the poll key to be used as a reset key",
        method=REST__GET,
        arguments=[
            SwaggerArgumentResponse(
                name=KEY__reset_or_poll_key,
                description="Either an reset or reset poll key",
                arg_type=str,
                example=[UUID__invite]
            ),
            SwaggerArgumentResponse(
                name=KEY__reset_code,
                description="An 8 length reset code received only via email. Used only with the reset poll key and valid only for 15 minutes. "
                "The reset key in the email is valid much longer (2 hours default)",
                arg_type=str,
                example=["7UYJ1TVX"],
                required=False,
                condition="Is this the reset poll key"
            )
        ],
        response=[
            SwaggerResponse(
                description="Reset status enumeration",
                response=SwaggerArgumentResponse(
                    name=KEY__reset_key_status,
                    description="An enumeration of the statuses of a reset key. 0->New reset key, 1->Reset process started, "
                    "2->Reset process finished ",
                    arg_type=int,
                    example=[0]
                )
            ),
            SwaggerFlatResponse(
                description=ERR__too_many_code_attempts,
                code=HTTPStatus.LOCKED,
                body=ERR__too_many_code_attempts
            )
        ]
    )
)

DOCUMENTATION__reset_password_with_invite = SwaggerDocumentation(
    tags="Account",
    # The security is in the invite key. User has not signed up yet so cannot get an oauth token
    security=False,
    methods=SwaggerMethod(
        name="Reset password with key",
        description="Resets password using either the key fetched either from internal methods or from the email",
        method=REST__POST,
        body=[
            ARG_RES__reset_key,
            ARG_RES__password
        ],
        response=[
            SwaggerResponse(
                description="The email address of the user that has had their password reset it's password",
                response=[ARG_RES__email, ARG_RES__parameters]
            )
        ]
    )
)

DOCUMENTATION__password = SwaggerDocumentation(
    tags="Account",
    methods=SwaggerMethod(
        method=REST__POST,
        name="Change password",
        description="Will change the logged in users password to the new one",
        arguments=[
            rename_arg(ARG_RES__username, KEY__old_password, "The old password for the user"),
            ARG_RES__password
        ],
        response=RES__oauth_token
    )
)

DOCUMENTATION__public_user = SwaggerDocumentation(
    tags="Public Users",
    security=False,
    methods=SwaggerMethod(
        method=REST__GET,
        name="Fetch public user",
        description="Fetches the public user",
        arguments=[
            set_required(ARG_RES__application, new_description="The application with which the user is associated"),
            ARG_RES__configuration
        ],
        response=SwaggerResponse(
            description="Credentials for the public user",
            response=[ARG_RES__username, ARG_RES__password]
        )
    )
)

DOCUMENTATION__submit = SwaggerDocumentation(
    tags="Queries",
    methods=SwaggerMethod(
        name="Execute JAAQL query",
        description="Executes a JAAQL query which is either a single SQL query or a list of queries. Returns results",
        method=REST__POST,
        arguments=ARG_RESP__allow_all,
        response=RES__allow_all,
        parallel_verification=True
    )
)

ARG_RES__renderable_document = [
    SwaggerArgumentResponse(
        name=KEY__attachment_name,
        description="The name of the renderable document in the database",
        arg_type=str,
        example=["my_pdf_template"]
    ),
    SwaggerArgumentResponse(
        name=KEY__parameters,
        description="Any parameters to pass to the url as http GET parameters",
        arg_type=ARG_RESP__allow_all
    ),
    set_nullable(ARG_RES__application, "Not required if being sent in an email as this is specified in the email"),
    set_nullable(ARG_RES__configuration, "Not required if being sent in an email as this is specified in the email"),
]

ARG_RES__attachments_for_send = SwaggerArgumentResponse(
    name=KEY__attachments,
    description="Any email attachments, if supplied. Uses server side template rendering",
    arg_type=SwaggerList(*ARG_RES__renderable_document),
    required=False,
    condition="If emails are attached"
)

ARG_RES__email_recipient = SwaggerArgumentResponse(
    name=KEY__recipient,
    description="The recipient of the email, null if sent to self",
    required=False,
    condition="Is the email being sent to someone else",
    arg_type=str,
    example=["my_manager", "client_a"]
)

ARG_RES__email_base = [
    ARG_RES__email_template,
    ARG_RES__email_recipient
]

DOCUMENTATION__email = SwaggerDocumentation(
    tags="Emails",
    methods=[
        SwaggerMethod(
            name="Send email",
            description="Sends an email template",
            method=REST__POST,
            body=[
                ARG_RES__application,
                ARG_RES__configuration,
                ARG_RES__parameters,
                ARG_RES__attachments_for_send
            ] + ARG_RES__email_base
        )
    ]
)

EXAMPLE__document_id = "b47dc954-d608-4e1b-8a8c-d8b754ee554b"

ARG_RES__document_id = SwaggerArgumentResponse(
    name=KEY__document_id,
    description="A document that can be used to fetch the document",
    arg_type=str,
    example=EXAMPLE__document_id
)

DOCUMENTATION__document = SwaggerDocumentation(
    tags="Documents",
    methods=[
        SwaggerMethod(
            name="Triggers a document render",
            description="Triggers a render of a document which can then be downloaded. Document is available for 5 minutes after the document has "
            "been rendered",
            method=REST__POST,
            body=ARG_RES__renderable_document + [
                SwaggerArgumentResponse(
                    name=KEY__create_file,
                    description="Whether or not to create the file. You will then be provided with a URL when it is ready which can be downloaded "
                                "from. Otherwise you will be sent back a boolean",
                    arg_type=bool
                )
            ],
            response=SwaggerResponse(
                description="A document id",
                response=ARG_RES__document_id
            )
        ),
        SwaggerMethod(
            name="Download document",
            description="Downloads the document. Can also be used as a polling endpoint to see if the document is ready",
            method=REST__GET,
            arguments=ARG_RES__document_id,
            response=[
                SwaggerFlatResponse(
                    description="A link to the raw file data. This URL is called with GET and no security parameters. Can only be called once",
                    body="https://www.jaaql.io/api/rendered_documents/" + EXAMPLE__document_id + ".pdf"
                ),
                SwaggerFlatResponse(
                    description="The url to the document. Will be deleted after 5 minutes",
                    code=HTTPStatus.CREATED,
                    body="https://www.jaaql.io/rendered_documents/" + EXAMPLE__document_id + ".pdf"
                ),
                SwaggerFlatResponse(
                    description="Document still rendering",
                    code=HTTP_STATUS__too_early,
                    body=ERR__document_still_rendering
                ),
                SwaggerFlatResponse(
                    description="Document id not found. Either expired or did not exist",
                    body=ERR__document_id_not_found,
                    code=HTTPStatus.NOT_FOUND
                )
            ]
        )
    ]
)

DOCUMENTATION__rendered_document = SwaggerDocumentation(
    tags="Documents",
    security=False,
    methods=SwaggerMethod(
        name="Stream rendered document",
        description="Streams a rendered document as a downlaod",
        method=REST__GET,
        arguments=[ARG_RES__document_id, SwaggerArgumentResponse(
            name=KEY__as_attachment,
            description="Whether in the browser the 'Content-Disposition' header should be set as attachment",
            arg_type=bool,
            required=False,
            condition="Defaults to false"
        )],
        response=SwaggerFlatResponse(
            description="The raw file data. Cannot be re-downloaded after this",
            body=BODY__file
        )
    )
)

DOCUMENTATION__drop_email_account = SwaggerDocumentation(
    tags="Emails",
    methods=SwaggerMethod(
        name="Drop email account",
        description="Drops the email account",
        method=REST__DELETE,
        arguments=SwaggerArgumentResponse(
            name=KEY__name,
            description="The name of the email account",
            arg_type=str,
            example=["noreply"]
        )
    )
)

ARG_RES__database_name = SwaggerArgumentResponse(
    name=KEY__name,
    description="The name of the database",
    arg_type=str,
    example=["invoicing_live"]
)



from setuptools import find_packages, setup
from jaaql.constants import VERSION

REQUIREMENTS = [i.strip().replace("==", "~=") for i in open("requirements.txt").readlines()]

setup(
    name='jaaql-middleware-python',
    packages=find_packages(include=['jaaql.*', 'jaaql']),
    version=VERSION,
    url='https://github.com/JAAQL/JAAQL-middleware-python',
    description='The jaaql package, allowing for rapid development and deployment of RESTful HTTP applications',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Software Quality Measurement and Improvement bv',
    author_email="aaron.tasker@sqmi.nl",
    license='Mozilla Public License Version 2.0 with Commons Clause',
    install_requires=REQUIREMENTS,
    package_data={'': ['config/*.ini', 'scripts/*.sql', 'migrations/*.sql', 'scripts/*.html']},
)



import requests
from os.path import join, exists, dirname
from datetime import datetime
import os
import glob
from jaaql.constants import DIR__config, FILE__config, CONFIG_KEY__server, CONFIG_KEY_SERVER__port, ENVIRON__install_path, PORT__ems
from jaaql.config_constants import *
from jaaql.db.db_interface import DBInterface
from jaaql.db.db_utils import create_interface
from jaaql.constants import VAULT_KEY__jaaql_lookup_connection
import configparser
import time
import urllib.parse
from jaaql.utilities.utils_no_project_imports import time_delta_ms  # Do not delete, relied on by others

PATH__migrations = "migrations"


class Profiler:
    def __init__(self, profile_id, do_profiling: bool):
        self.profile_id = profile_id
        self.do_profiling = do_profiling
        self.cur_time = datetime.now()

    def perform_profile(self, description: str):
        if not self.do_profiling:
            return

        cur_time = str(time_delta_ms(self.cur_time, datetime.now()))
        print("PROFILING: " + str(self.profile_id) + " " + cur_time + " " + description)
        self.cur_time = datetime.now()

    def copy(self):
        return Profiler(self.profile_id, self.do_profiling)


def get_jaaql_root() -> str:
    return dirname(dirname(__file__))


def load_root_config(is_gunicorn):
    config_root = get_jaaql_root()
    if is_gunicorn:
        config_root = "/JAAQL-middleware-python/jaaql"

    config = configparser.ConfigParser()
    config.sections()
    config_path = join(config_root, DIR__config, FILE__config)
    if not exists(config_path):
        raise Exception("Could not find config. Please check working directory has access to '" + config_path + "'")
    config.read(config_path)
    return {s: dict(config.items(s)) for s in config.sections()}


def load_config(is_gunicorn, config_path=None):
    config = load_root_config(is_gunicorn)
    if config_path is None:
        config_path = join(DIR__config, FILE__config)
    if exists(config_path):
        override_config = configparser.ConfigParser()
        override_config.sections()
        override_config.read(config_path)
        override_config = {s: dict(override_config.items(s)) for s in override_config.sections()}
        for each, _ in override_config.items():
            if each in config:
                for sub_each, val in override_config[each].items():
                    if sub_each in config[each]:
                        config[each][sub_each] = val
            else:
                config[each] = override_config[each]
    return config


def loc(data: [dict], key: str, val: str):
    return [row for row in data if row[key] == val]


def load_email_templates():
    templates_locs = glob.glob("templates/*.html")
    templates = {}
    for template in templates_locs:
        template = template.replace("\\", "/")
        with open(template, "r") as f:
            templates[".".join(template.split("templates/")[1].split(".")[:-1])] = f.read()
    return templates


def get_base_url(config, is_gunicorn: bool):
    if is_gunicorn:
        return "http+unix://" + urllib.parse.quote(os.environ.get(ENVIRON__install_path) + "/jaaql.sock", safe='')
    else:
        return "http://127.0.0.1:" + str(int(config[CONFIG_KEY__server][CONFIG_KEY_SERVER__port]))


def await_jaaql_bootup(config, is_gunicorn: bool):
    base_url = get_base_url(config, is_gunicorn)
    while True:
        try:
            return requests.get(base_url + "/internal/is_installed").status_code
        except:
            pass
        time.sleep(5)


def await_jaaql_installation(config, is_gunicorn: bool):
    base_url = get_base_url(config, is_gunicorn)
    while True:
        try:
            if requests.get(base_url + "/internal/is_installed").status_code == 200:
                break
        except:
            pass
        time.sleep(5)


def get_external_url(config):
    return config[CONFIG_KEY__swagger][CONFIG_KEY_SWAGGER__url]


def await_ems_startup():
    while True:
        try:
            if requests.get("http://127.0.0.1:" + str(PORT__ems) + "/").status_code == 200:
                break
        except:
            pass
        time.sleep(5)


def get_jaaql_connection(config, vault):
    jaaql_uri = vault.get_obj(VAULT_KEY__jaaql_lookup_connection)
    address, port, db, username, password = DBInterface.fracture_uri(jaaql_uri)
    return create_interface(config, address, port, db, username, password)

def get_db_connection_as_jaaql(config, vault, db: str):
    jaaql_uri = vault.get_obj(VAULT_KEY__jaaql_lookup_connection)
    address, port, _, username, password = DBInterface.fracture_uri(jaaql_uri)
    return create_interface(config, address, port, db, username, password)



from datetime import datetime
import os
from jaaql.constants import ENVIRON__install_path
from os.path import join
import requests
from typing import Union


def time_delta_ms(start_time: datetime, end_time: datetime) -> int:
    return int(round((end_time - start_time).total_seconds() * 1000))


def load_artifact(is_container: bool, artifact_base_uri: str, app_relative_path: str):
    if app_relative_path is None:
        return None

    if check_allowable_file_path(app_relative_path):
        raise Exception("Database has been tampered with! Cannot send email")
    if not app_relative_path.endswith(".html"):
        app_relative_path = app_relative_path + ".html"
    if not artifact_base_uri.startswith("https://") and not artifact_base_uri.startswith("http://"):
        if is_container:
            if check_allowable_file_path(artifact_base_uri):
                print(artifact_base_uri)
                raise Exception("Illegal artifact base url")
        if artifact_base_uri.startswith("file:///"):
            return open(join(artifact_base_uri.split("file:///")[1].replace("%20", " "), app_relative_path), "r").read()
        else:
            return open(join(os.environ[ENVIRON__install_path], "www", artifact_base_uri, app_relative_path), "r").read()
    else:
        splitter = "" if artifact_base_uri.endswith("/") else "/"
        return requests.get(artifact_base_uri + splitter + app_relative_path).text


def check_allowable_file_path(uri):
    return not isinstance(uri, str) or any([not letter.isalnum() and letter not in ['-', '_', '/'] for letter in uri])


def pull_from_dict(self, inputs: dict, keys: Union[list, str, dict]):
    if not isinstance(keys, list) and not isinstance(keys, dict):
        keys = [keys]
    if isinstance(keys, list):
        return {key: inputs[key] for key in keys}
    else:
        return {map_to: inputs[map_from if map_from is not None else map_to] for map_from, map_to in keys.items()}



from jaaql.db.db_interface import DBInterface
from jaaql.db.db_utils import execute_supplied_statement_singleton, execute_supplied_statement
from jaaql.mvc.queries import QUERY__load_application_configuration, QUERY__load_application_configurations
from jaaql.constants import KEY__application, KEY__configuration, KEY__artifact_base_uri, FILE__canned_queries
from jaaql.utilities.utils_no_project_imports import load_artifact
from jaaql.exceptions.http_status_exception import HttpStatusException
from requests.exceptions import RequestException
import json

ERR__invalid_query = "Could not find the canned query that you have specified! Could not find %s"


class CachedCannedQueryService:

    def __init__(self, is_container: bool, connection: DBInterface):
        self.canned_queries = {}
        # self.init_canned_queries(is_container, connection)

    def init_canned_queries(self, is_container: bool, connection: DBInterface):
        configs = execute_supplied_statement(connection, QUERY__load_application_configurations, as_objects=True)
        for config in configs:
            self.refresh_configuration(is_container, connection, config[KEY__application], config[KEY__configuration],
                                       config[KEY__artifact_base_uri])

    def get_canned_query(self, application: str, configuration: str, file: str, pos: int) -> str:
        if application not in self.canned_queries:
            raise HttpStatusException(ERR__invalid_query % "application")

        if configuration not in self.canned_queries[application]:
            raise HttpStatusException(ERR__invalid_query % "configuration")

        queries = self.canned_queries[application][configuration]

        if file not in queries:
            raise HttpStatusException(ERR__invalid_query % "file")

        if pos not in queries[file]:
            raise HttpStatusException(ERR__invalid_query % "pos")

        return queries[file][pos]

    def refresh_configuration(self, is_container: bool, connection: DBInterface, application: str, configuration: str,
                              config_resource_url: str = None):
        if config_resource_url is None:
            config_resource_url = execute_supplied_statement_singleton(connection, QUERY__load_application_configuration,
                                                                       {KEY__application: application, KEY__configuration: configuration},
                                                                       as_objects=True)[KEY__artifact_base_uri]

        canned_queries = None
        try:
            canned_queries = load_artifact(is_container, config_resource_url, FILE__canned_queries)
            json.loads(canned_queries)
        except FileNotFoundError:
            pass  # No canned queries for app. This is _okay_
        except RequestException:
            pass  # Same as above

        if application not in canned_queries:
            self.canned_queries[application] = {}

        if configuration not in canned_queries[application]:
            self.canned_queries[application][configuration] = {}

        self.canned_queries[application][configuration] = canned_queries

        # TODO broadcast this to other workers



CONFIG_KEY__security = "SECURITY"
CONFIG_KEY_SECURITY__mfa_label = "mfa_label"
CONFIG_KEY_SECURITY__force_mfa = "force_mfa"
CONFIG_KEY_SECURITY__do_audit = "do_audit"
CONFIG_KEY__swagger = "SWAGGER"
CONFIG_KEY_SWAGGER__url = "url"
CONFIG_KEY__server = "SERVER"
CONFIG_KEY_SERVER__port = "port"
CONFIG_KEY_SECURITY__mfa_issuer = "mfa_issuer"
CONFIG_KEY_SECURITY__token_expiry_ms = "token_expiry_ms"
CONFIG_KEY_SECURITY__token_refresh_expiry_ms = "token_refresh_expiry_ms"